// Copyright (c) 2024, Jay Shah, Ganesh Bikshandi, Ying Zhang, Vijay Thakkar, Pradeep Ramani, Tri Dao.
// Splitting the different template instantiations to different files to speed up compilation.
// This file is auto-generated. See "generate_kernels.py"

#include "flash_fwd_hdim64_fp16_sm90.cu"
#include "flash_fwd_hdim96_fp16_sm90.cu"
#include "flash_fwd_hdim128_fp16_sm90.cu"
#include "flash_fwd_hdim192_fp16_sm90.cu"
#include "flash_fwd_hdim256_fp16_sm90.cu"